@@ -64,11 +64,6 @@ sub run {
     $host->set('Log Stream', $self->get('Log Stream'));
     $self->set('Host', $host);
 
-    my $db = Sbuild::DB::Client->new($self->get('Config'));
-    $db->set('Log Stream', $self->get('Log Stream'));
-    $self->set('DB', $db);
-
-    my @distlist = qw(oldstable-security stable-security testing-security stable testing unstable);
     my $my_binary = $0;
     $my_binary = cwd . "/" . $my_binary if $my_binary !~ m,^/,;
     $self->set('MY_BINARY', $my_binary);
@@ -95,8 +90,8 @@ sub run {
 	}
     }
 
-    if (@{$self->get_conf('TAKE_FROM_DISTS')}) {
-	die "take_from_dists is empty, aborting.";
+    if (!@{$self->get_conf('DISTRIBUTIONS')}) {
+	die "distribution list is empty, aborting.";
     }
 
     if (!$self->get_conf('NO_DETACH')) {
@@ -118,49 +113,50 @@ sub run {
     undef $ENV{'DISPLAY'};
 
 # the main loop
-    my $dist;
   MAINLOOP:
     while( 1 ) {
 	$self->check_restart();
 	$self->read_config();
-	$self->check_ssh_master();
 
 	my $done = 0;
 	my $thisdone;
 	my %binNMUlog;
 	do {
 	    $thisdone = 0;
-	    foreach $dist (@{$self->get_conf('TAKE_FROM_DISTS')}) {
+	    foreach my $dist_config (@{$self->get_conf('DISTRIBUTIONS')}) {
 		$self->check_restart();
 		$self->read_config();
-		my @redo = $$self->get_from_REDO( $dist, %binNMUlog );
-		next if !@redo;
-		$self->do_build( $dist, %binNMUlog, @redo );
+		my $pkg_ver = $self->get_from_REDO( $dist_config, \%binNMUlog );
+		next if !defined($pkg_ver);
+		$self->do_build( $dist_config, \%binNMUlog, $pkg_ver);
 		++$done;
 		++$thisdone;
 	    }
 	} while( $thisdone );
 
-	foreach $dist (@{$self->get_conf('TAKE_FROM_DISTS')}) {
+	foreach my $dist_config (@{$self->get_conf('DISTRIBUTIONS')}) {
 	    $self->check_restart();
 	    $self->read_config();
+	    $self->check_ssh_master($dist_config);
+	    my $dist_name = $dist_config->get('DIST_NAME');
 	    my %givenback = $self->read_givenback();
-	    my $pipe = $self->get('DB')->pipe_query('--list=needs-build',
-						    "--dist=$dist");
+		my $db = $self->get_db_handle($dist_config);
+	    my $pipe = $db->pipe_query(
+		'--list=needs-build',
+		'--dist=' . $dist_name);
 	    if (!$pipe) {
 		$self->log("Can't spawn wanna-build --list=needs-build: $!\n");
 		next MAINLOOP;
 	    }
 
-	    my(@todo, $total, $nonex, @lowprio_todo, $max_build);
-	    $max_build = $self->get_conf('MAX_BUILD');
+	    my(@todo, $total, $nonex, @lowprio_todo);
 	    while( <$pipe> ) {
-		my $socket = $self->get_conf('WANNA_BUILD_SSH_SOCKET');
+		my $socket = $dist_config->get('WANNA_BUILD_SSH_SOCKET');
 		if ($socket &&
 		    (/^Couldn't connect to $socket: Connection refused[\r]?$/ ||
 		     /^Control socket connect\($socket\): Connection refused[\r]?$/)) {
 		    unlink($socket);
-		    $self->check_ssh_master();
+		    $self->check_ssh_master($dist_config);
 		}
 		elsif (/^Total (\d+) package/) {
 		    $total = $1;
@@ -170,52 +166,42 @@ sub run {
 		    $nonex = 1;
 		}
 		next if $nonex;
-		next if @todo >= $max_build;
+		next if @todo >= 1; #we only want one!
 		my @line = (split( /\s+/, $_));
 		my $pv = $line[0];
-		my $no_build_regex = $self->get_conf('NO_BUILD_REGEX');
-		my $build_regex = $self->get_conf('BUILD_REGEX');
+		my $no_build_regex = $dist_config->get('NO_BUILD_REGEX');
+		my $build_regex = $dist_config->get('BUILD_REGEX');
 		next if $no_build_regex && $pv =~ m,$no_build_regex,;
 		next if $build_regex && $pv !~ m,$build_regex,;
 		$pv =~ s,^.*/,,;
 		my $p;
 		($p = $pv) =~ s/_.*$//;
-		next if isin( $p, @{$self->get_conf('NO_AUTO_BUILD')} );
+		next if isin( $p, @{$dist_config->get('NO_AUTO_BUILD')} );
 		next if $givenback{$pv};
-		if (isin( $p, @{$self->get_conf('WEAK_NO_AUTO_BUILD')} )) {
+		if (isin( $p, @{$dist_config->get('WEAK_NO_AUTO_BUILD')} )) {
 		    push( @lowprio_todo, $pv );
 		    next;
 		}
-		if ($line[1] =~ /:binNMU/) {
-		    $max_build = 1;
-		    @todo = ();
-		}
 		push( @todo, $pv );
 	    }
 	    close( $pipe );
 	    next if $nonex;
 	    if ($?) {
-		$self->log("wanna-build --list=needs-build --dist=$dist failed; status ",
+		$self->log("wanna-build --list=needs-build --dist=${dist_name} failed; status ",
 			   exitstatus($?), "\n");
 		next;
 	    }
-	    $self->log("$dist: total $total packages to build.\n") if $total;
-	    if ($total && $self->get_conf('SECONDARY_DAEMON_THRESHOLD') &&
-		$total < $self->get_conf('SECONDARY_DAEMON_THRESHOLD')) {
-		$self->log("Not enough packages to build -- ".
-			   "secondary daemon not starting\n");
-		next;
-	    }
+	    $self->log("${dist_name}: total $total packages to build.\n") if defined($total);
 
 	    # Build weak_no_auto packages before the next dist
 	    if (!@todo && @lowprio_todo) {
-		push ( @todo, @lowprio_todo );
+		push @todo, $lowprio_todo[0];
 	    }
 
 	    next if !@todo;
-	    @todo = $self->do_wanna_build( $dist, %binNMUlog, @todo );
+	    @todo = $self->do_wanna_build( $dist_config, \%binNMUlog, @todo );
 	    next if !@todo;
-	    $self->do_build( $dist, %binNMUlog, @todo );
+	    $self->do_build( $dist_config, \%binNMUlog, $todo[0] );
 	    ++$done;
 	    last;
 	}
@@ -236,9 +222,9 @@ sub run {
 
 sub get_from_REDO {
     my $self = shift;
-    my $wanted_dist = shift;
+    my $wanted_dist_config = shift;
     my $binNMUlog = shift;
-    my @redo = ();
+    my $pkg_ver = undef;
     local( *F );
 
     lock_file( "REDO" );
@@ -256,27 +242,20 @@ sub get_from_REDO {
 		   "Raw contents:\n@lines\n");
 	goto end;
     }
-    my $max_build = $self->get_conf('MAX_BUILD');
     foreach (@lines) {
 	if (!/^(\S+)\s+(\S+)(?:\s*|\s+(\d+)\s+(\S.*))?$/) {
 	    $self->log("Ignoring/deleting bad line in REDO: $_");
 	    next;
 	}
 	my($pkg, $dist, $binNMUver, $changelog) = ($1, $2, $3, $4);
-	if ($dist eq $wanted_dist && @redo < $max_build) {
+	if ($dist eq $wanted_dist_config->get('DIST_NAME') && !defined($pkg_ver)) {
 	    if (defined $binNMUver) {
-		if (scalar(@redo) == 0) {
-		    $binNMUlog->{$pkg} = $changelog;
-		    push( @redo, "!$binNMUver!$pkg" );
-		} else {
-		    print F $_;
-		}
-		$max_build = scalar(@redo);
+		$binNMUlog->{$pkg} = $changelog;
+		$pkg_ver = "!$binNMUver!$pkg";
 	    } else {
-		push( @redo, $pkg );
+		$pkg_ver = $pkg;
 	    }
-	}
-	else {
+	} else {
 	    print F $_;
 	}
     }
@@ -285,7 +264,24 @@ sub get_from_REDO {
   end:
     unlock_file( "REDO" );
     $self->unblock_signals();
-    return @redo;
+    return $pkg_ver;
+}
+
+sub add_given_back ($$) {
+    my $self = shift;
+    my $pkg_ver = shift;
+
+    local( *F );
+    lock_file("SBUILD-GIVEN-BACK", 0);
+
+    if (open( F, ">>SBUILD-GIVEN-BACK" )) {
+	print F $pkg_ver . " " . time() . "\n";
+	close( F );
+    } else {
+	$self->log("Can't open SBUILD-GIVEN-BACK: $!\n");
+    }
+
+    unlock_file("SBUILD-GIVEN-BACK");
 }
 
 sub read_givenback {
@@ -325,15 +321,19 @@ sub read_givenback {
 sub do_wanna_build {
     my $self = shift;
 
-    my $dist = shift;
+    my $dist_config = shift;
     my $binNMUlog = shift;
     my @output = ();
     my $n = 0;
 
     $self->block_signals();
 
-    my $pipe = $self->get('DB')->pipe_query('-v', "--dist=$dist", @_);
-    if (!$pipe) {
+    my $db = $self->get_db_handle($dist_config);
+    my $pipe = $db->pipe_query(
+	'-v', 
+	'--dist=' . $dist_config->get('DIST_NAME'),
+       	@_);
+    if ($pipe) {
 	while( <$pipe> ) {
 	    next if /^wanna-build Revision/;
 	    if (/^(\S+):\s*ok/) {
@@ -352,7 +352,7 @@ sub do_wanna_build {
 		my $pkg = $1;
 		++$n;
 		if ($self->get_conf('SHOULD_BUILD_MSGS')) {
-		    $self->handle_prevfailed( $dist, grep( /^\Q$pkg\E_/, @_ ) );
+		    $self->handle_prevfailed( $dist_config, grep( /^\Q$pkg\E_/, @_ ) );
 		} else {
 		    push( @output, grep( /^\Q$pkg\E_/, @_ ) );
 		}
@@ -391,10 +391,10 @@ sub do_wanna_build {
 
 sub do_build {
     my $self = shift;
-    my $dist = shift;
+    my $dist_config = shift;
     my $binNMUlog = shift;
+    my $pkg_ver = shift;
 
-    return if !@_;
     my $free_space;
 
     while (($free_space = df(".")) < $self->get_conf('MIN_FREE_SPACE')) {
@@ -405,42 +405,45 @@ sub do_build {
 	$self->write_stats("idle-time", $idle_end_time - $idle_start_time);
     }
 
-    $self->log("Starting build (dist=$dist) of:\n@_\n");
-    $self->write_stats("builds", scalar(@_));
+    $self->log("Starting build (dist=" . $dist_config->get('DIST_NAME') . ") of $pkg_ver\n");
+    $self->write_stats("builds", 1);
     my $binNMUver;
 
     my @sbuild_args = ( 'nice', '-n', $self->get_conf('NICE_LEVEL'), 'sbuild',
 			'--apt-update',
 			'--batch',
 			"--stats-dir=" . $self->get_conf('HOME') . "/stats",
-			"--dist=$dist" );
-    my $sbuild_gb = '--auto-give-back';
-    if ($self->get_conf('WANNA_BUILD_SSH_CMD')) {
-	$sbuild_gb .= "=";
-	$sbuild_gb .= $self->get_conf('WANNA_BUILD_SSH_SOCKET') . "\@"
-	    if $self->get_conf('WANNA_BUILD_SSH_SOCKET');
-	$sbuild_gb .= $self->get_conf('WANNA_BUILD_DB_USER') . "\@"
-	    if $self->get_conf('WANNA_BUILD_DB_USER');
-	$sbuild_gb .= $self->get_conf('WANNA_BUILD_SSH_USER') ."\@" if $self->get_conf('WANNA_BUILD_SSH_USER');
-	$sbuild_gb .= $self->get_conf('WANNA_BUILD_SSH_HOST');
-    } else {
-	# Otherwise newer sbuild will take the package name as an --auto-give-back
-	# parameter (changed from regexp to GetOpt::Long parsing)
-	$sbuild_gb .= "=yes"
+			"--dist=" . $dist_config->get('DIST_NAME') );
+    #multi-archive-buildd keeps the mailto configuration in the builddrc, so
+    #this needs to be passed over to sbuild. If the buildd config doesn't have
+    #it, we hope that the address is configured in .sbuildrc and the right one:
+    if ($dist_config->get('LOGS_MAILED_TO')) {
+	push @sbuild_args, '--mail-log-to=' . $dist_config->get('LOGS_MAILED_TO');
     }
-    push ( @sbuild_args, $sbuild_gb );
-    push ( @sbuild_args, "--database=" . $self->get_conf('WANNA_BUILD_DB_NAME') )
-	if $self->get_conf('WANNA_BUILD_DB_NAME');
+    #Some distributions (bpo, experimental) require a more complex dep resolver.
+    #Ask sbuild to use another build-dep resolver if the config says so:
+    if ($dist_config->get('BUILD_DEP_RESOLVER')) {
+	push @sbuild_args, '--build-dep-resolver=' . $dist_config->get('BUILD_DEP_RESOLVER');
+    }
+    push ( @sbuild_args, "--arch=" . $dist_config->get('BUILT_ARCHITECTURE') )
+	if $dist_config->get('BUILT_ARCHITECTURE');
+    push ( @sbuild_args, "--chroot=" . $dist_config->get('SBUILD_CHROOT') )
+	if $dist_config->get('SBUILD_CHROOT');
 
-    if (scalar(@_) == 1 and $_[0] =~ s/^!(\d+)!//) {
+
+    if ($pkg_ver =~ s/^!(\d+)!//) {
 	$binNMUver = $1;
 
-	push ( @sbuild_args, "--binNMU=$binNMUver", "--make-binNMU=" . $binNMUlog->{$_[0]});
+	push ( @sbuild_args, "--binNMU=$binNMUver", "--make-binNMU=" . $binNMUlog->{$pkg_ver});
     }
-    $self->log("command line: @sbuild_args @_\n");
+    push @sbuild_args, $pkg_ver;
+    $self->log("command line: @sbuild_args\n");
 
-    if (($main::sbuild_pid = fork) == 0) {
-	{ exec (@sbuild_args, @_) };
+    $main::sbuild_pid = open(SBUILD_OUT, "-|");
+
+    #We're childish, so call sbuild:
+    if ($main::sbuild_pid == 0) {
+	{ exec (@sbuild_args) };
 	$self->log("Cannot execute sbuild: $!\n");
 	exit(64);
     }
@@ -449,6 +452,17 @@ sub do_build {
 	$self->log("Cannot fork for sbuild: $!\n");
 	goto failed;
     }
+
+    #We want to collect the first few lines of sbuild output:
+    my ($sbuild_output_line_count, @sbuild_output_buffer) = (0, ());
+    while (<SBUILD_OUT>) {
+	#5 lines are enough:
+	if (++$sbuild_output_line_count < 5) {
+	    push @sbuild_output_buffer, $_;
+	}
+    }
+
+    #We got enough output, now just wait for sbuild to die:
     my $rc;
     while (($rc = wait) != $main::sbuild_pid) {
 	if ($rc == -1) {
@@ -459,53 +473,63 @@ sub do_build {
 	    $self->log("wait for sbuild: returned unexpected pid $rc\n");
 	}
     }
+    my $sbuild_exit_code = $?;
     undef $main::sbuild_pid;
+    close(SBUILD_OUT);
 
-    if ($?) {
-	$self->log("sbuild failed with status ".exitstatus($?)."\n");
-      failed:
-	if (-f "SBUILD-REDO-DUMPED") {
-	    $self->log("Found SBUILD-REDO-DUMPED; sbuild already dumped ",
-		       "pkgs which need rebuiling/\n");
-	    local( *F );
-	    my $n = 0;
-	    open( F, "<REDO" );
-	    while( <F> ) { ++$n; }
-	    close( F );
-	    $self->write_stats("builds", -$n);
+    #Process sbuild's results:
+    my $db = $self->get_db_handle($dist_config);
+    my $failed = 1;
+    my $giveback = 1;
+
+    if (WIFEXITED($sbuild_exit_code)) {
+	my $status = WEXITSTATUS($sbuild_exit_code);
+
+	if ($status == 0) {
+	    $failed = 0;
+	    $giveback = 0;
+	    $self->log("sbuild of $pkg_ver succeeded -- marking as built in wanna-build\n");
+	    $db->run_query('--built', '--dist=' . $dist_config->get('DIST_NAME'), $pkg_ver);
+	} elsif ($status ==  2) {
+	    $giveback = 0;
+	    $self->log("sbuild of $pkg_ver failed with status $status (build failed) -- marking as attempted in wanna-build\n");
+	    $db->run_query('--attempted', '--dist=' . $dist_config->get('DIST_NAME'), $pkg_ver);
+	    $self->write_stats("failed", 1);
+	} else {
+	    $self->log("sbuild of $pkg_ver failed with status $status (local problem) -- giving back\n");
 	}
-	elsif (-f "SBUILD-FINISHED") {
-	    my @finished = $self->read_FINISHED();
-	    $self->log("sbuild has already finished:\n@finished\n");
-	    my @unfinished;
-	    for (@_) {
-		push( @unfinished, $_ ) if !isin( $_, @finished );
-	    }
-	    $self->log("Adding rest to REDO:\n@unfinished\n");
-	    $self->append_to_REDO( $dist, '', @unfinished );
-	    $self->write_stats("builds", -scalar(@unfinished));
-	}
-	else {
-	    if (defined $binNMUver) {
-		$self->log("Assuming binNMU failed and adding to REDO:\n@_\n");
-		$self->append_to_REDO( $dist, "$binNMUver $binNMUlog->{$_[0]}", @_ );
-	    } else {
-		$self->log("Assuming all packages unbuilt and adding to REDO:\n@_\n");
-		$self->append_to_REDO( $dist, '', @_ );
-	    }
-	    $self->write_stats("builds", -scalar(@_));
+    } elsif (WIFSIGNALED($sbuild_exit_code)) {
+	my $sig = WTERMSIG($sbuild_exit_code);
+	$self->log("sbuild of $pkg_ver failed with signal $sig (local problem) -- giving back\n");
+    } else {
+	$self->log("sbuild of $pkg_ver failed with unknown reason (local problem) -- giving back\n");
+    }
+
+    if ($giveback) {
+	$db->run_query('--give-back', '--dist=' . $dist_config->get('DIST_NAME'), $pkg_ver);
+	$self->add_given_back($pkg_ver);
+	$self->write_stats("give-back", 1);
+    }
+
+    # Check if we encountered some local error to stop further building
+    if ($giveback) {
+	delete $binNMUlog->{$pkg_ver} if defined $binNMUver;
+
+	if (!defined $main::sbuild_fails) {
+	    $main::sbuild_fails = 0;
 	}
 
-	delete $binNMUlog->{$_[0]} if defined $binNMUver;
+	$main::sbuild_fails++;
 
-	if (++$main::sbuild_fails > 2) {
+	if ($main::sbuild_fails > 2) {
 	    $self->log("sbuild now failed $main::sbuild_fails times in ".
 		       "a row; going to sleep\n");
 	    send_mail( $self->get_conf('ADMIN_MAIL'),
 		       "Repeated mess with sbuild",
 		       <<EOF );
 The execution of sbuild now failed for $main::sbuild_fails times.
-Something must be wrong here...
+These are the first $sbuild_output_line_count lines of the last failed sbuild call:
+@sbuild_output_buffer
 
 The daemon is going to sleep for 1 hour, or can be restarted with SIGUSR2.
 EOF
@@ -522,23 +546,29 @@ EOF
 	}
     }
     else {
+	# Either a build success or an attempted build will cause the
+	# counter to reset.
 	$main::sbuild_fails = 0;
     }
-    unlink "SBUILD-REDO-DUMPED" if -f "SBUILD-REDO-DUMPED";
     $self->log("Build finished.\n");
 }
 
 sub handle_prevfailed {
     my $self = shift;
-    my $dist = shift;
+    my $dist_config = shift;
     my $pkgv = shift;
 
+    my $dist_name = $dist_config->get('DIST_NAME');
     my( $pkg, $fail_msg, $changelog);
 
     $self->log("$pkgv previously failed -- asking admin first\n");
     ($pkg = $pkgv) =~ s/_.*$//;
 
-    my $pipe = $self->get('DB')->pipe_query('--info', "--dist=$dist", $pkg);
+    my $db = $self->get_db_handle($dist_config);
+    my $pipe = $db->pipe_query(
+	'--info',
+       	'--dist=' . $dist_name,
+       	$pkg);
     if (!$pipe) {
 	$self->log("Can't run wanna-build: $!\n");
 	return;
@@ -552,22 +582,11 @@ sub handle_prevfailed {
 	return;
     }
 
-    {
-	local $SIG{'ALRM'} = sub ($) { die "Timeout!\n" };
-	eval { $changelog = $self->get_changelog( $dist, $pkgv ) };
-    }
-    $changelog = "ERROR: FTP timeout" if $@;
-
     send_mail( $self->get_conf('ADMIN_MAIL'),
-	       "Should I build $pkgv (dist=$dist)?",
+	       "Should I build $pkgv (dist=${dist_name})?",
 	       "The package $pkg failed to build in a previous version. ".
 	       "The fail\n".
 	       "messages are:\n\n$fail_msg\n".
-	       ($changelog !~ /^ERROR/ ?
-		"The changelog entry for the newest version is:\n\n".
-		"$changelog\n" :
-		"Sorry, the last changelog entry could not be extracted:\n".
-		"$changelog\n\n").
 	       "Should buildd try to build the new version, or should it ".
 	       "fail with the\n".
 	       "same messages again.? Please answer with 'build' (or 'ok'), ".
@@ -575,10 +594,16 @@ sub handle_prevfailed {
 }
 
 sub get_changelog {
+    # This method is currently broken.  It makes some assumptions about source
+    # layout that are no longer true.  Furthermore it tries fetching through
+    # the host instead of creating a session (which is necessary for snapshot-
+    # based chroots) and work in the chroot.
+
     my $self = shift;
-    my $dist = shift;
+    my $dist_config = shift;
     my $pkg = shift;
 
+    my $dist_name = $dist_config->get('DIST_NAME');
     my $changelog = "";
     my $analyze = "";
     my $chroot_apt_options;
@@ -592,9 +617,9 @@ sub get_changelog {
 
 retry:
     my @schroot = ($self->get_conf('SCHROOT'), '-c',
-		   "$dist-" . $self->get_conf('ARCH') . '-sbuild', '--');
+		   $dist_name . '-' . $self->get_conf('ARCH') . '-sbuild', '--');
     my @schroot_root = ($self->get_conf('SCHROOT'), '-c',
-			"$dist-" . $self->get_conf('ARCH') . '-sbuild',
+			$dist_name . '-' . $self->get_conf('ARCH') . '-sbuild',
 			'-u', 'root', '--');
     my $apt_get = $self->get_conf('APT_GET');
 
@@ -702,56 +727,6 @@ retry:
     return $changelog;
 }
 
-sub append_to_REDO {
-    my $self = shift;
-    my $dist = shift;
-    my $postfix = shift;
-
-    my @npkgs = @_;
-    my @pkgs = ();
-    my $pkg;
-    local( *F );
-
-    $self->block_signals();
-    lock_file( "REDO" );
-
-    if (open( F, "REDO" )) {
-	@pkgs = <F>;
-	close( F );
-    }
-
-    if (open( F, ">>REDO" )) {
-	foreach $pkg (@npkgs) {
-	    next if grep( /^\Q$pkg\E\s/, @pkgs );
-	    print F "$pkg ${dist}$postfix\n";
-	}
-	close( F );
-    }
-    else {
-	$self->log("Can't open REDO: $!\n");
-    }
-
-  unlock:
-    unlock_file( "REDO" );
-    $self->unblock_signals();
-}
-
-sub read_FINISHED {
-    my $self = shift;
-
-    local( *F );
-    my @pkgs;
-
-    if (!open( F, "<SBUILD-FINISHED" )) {
-	$self->log("Can't open SBUILD-FINISHED: $!\n");
-	return ();
-    }
-    chomp( @pkgs = <F> );
-    close( F );
-    unlink( "SBUILD-FINISHED" );
-    return @pkgs;
-}
-
 sub check_restart {
     my $self = shift;
     my @stats = stat( $self->get('MY_BINARY') );
@@ -765,7 +740,7 @@ sub check_restart {
 
     if ( -f $self->get_conf('HOME') . "/EXIT-DAEMON-PLEASE" ) {
 	unlink($self->get_conf('HOME') . "/EXIT-DAEMON-PLEASE");
-	&shutdown("NONE (flag file exit)");
+	$self->shutdown("NONE (flag file exit)");
     }
 }
 
@@ -783,29 +758,46 @@ sub unblock_signals {
 
 sub check_ssh_master {
     my $self = shift;
+    my $dist_config = shift;
 
-    return 1 if (!$self->get_conf('WANNA_BUILD_SSH_SOCKET'));
-    return 1 if ( -S $self->get_conf('WANNA_BUILD_SSH_SOCKET') );
+    my $ssh_socket = $dist_config->get('WANNA_BUILD_SSH_SOCKET');
 
-    if ($main::ssh_pid)
-    {
-	my $wpid = waitpid ( $main::ssh_pid, WNOHANG );
-	return 1 if ($wpid != -1 and $wpid != $main::ssh_pid);
+    return 1 if (!$ssh_socket);
+    return 1 if ( -S $ssh_socket );
+
+    my $ssh_master_pids = {};
+    if ($self->get('SSH_MASTER_PIDS')) {
+	$ssh_master_pids = $self->get('SSH_MASTER_PIDS');
+    } else {
+	$self->set('SSH_MASTER_PIDS', $ssh_master_pids);
     }
 
-    ($main::ssh_pid = fork)
-	or exec (@{$self->get_conf('WANNA_BUILD_SSH_CMD')}, "-MN");
+    if ($ssh_master_pids->{$ssh_socket})
+    {
+	my $wpid = waitpid ( $ssh_master_pids->{$ssh_socket}, WNOHANG );
+	return 1 if ($wpid != -1 and $wpid != $ssh_master_pids->{$ssh_socket});
+    }
 
-    if (!defined $main::ssh_pid) {
+    my $new_master_pid = fork;
+
+    #We are in the newly forked child:
+    if (defined($new_master_pid) && $new_master_pid == 0) {
+	exec (@{$dist_config->get('WANNA_BUILD_SSH_CMD')}, "-MN");
+    }
+
+    #We are the parent:
+    if (!defined $new_master_pid) {
 	$self->log("Cannot fork for ssh master: $!\n");
 	return 0;
     }
 
-    while ( ! -S $self->get_conf('WANNA_BUILD_SSH_SOCKET') )
+    $ssh_master_pids->{$ssh_socket} = $new_master_pid;
+
+    while ( ! -S $ssh_socket )
     {
 	sleep 1;
-	my $wpid = waitpid ( $main::ssh_pid, WNOHANG );
-	return 0 if ($wpid == -1 or $wpid == $main::ssh_pid);
+	my $wpid = waitpid ( $new_master_pid, WNOHANG );
+	return 0 if ($wpid == -1 or $wpid == $new_master_pid);
     }
     return 1;
 }
@@ -816,4 +808,41 @@ sub read_config {
     $self->get('Config')->read_config();
 }
 
+sub shutdown {
+    my $self = shift;
+    my $signame = shift;
+
+    $self->log("buildd ($$) received SIG$signame -- shutting down\n");
+
+    if ($self->get('SSH_MASTER_PIDS')) {
+	my $ssh_master_pids = $self->get('SSH_MASTER_PIDS');
+	for my $ssh_socket (keys %{$ssh_master_pids}) {
+	    my $master_pid = $ssh_master_pids->{$ssh_socket};
+	    kill ( 15, $master_pid );
+	    delete ( $ssh_master_pids->{$ssh_socket} );
+	}
+    }
+
+    if (defined $main::sbuild_pid) {
+	$self->log("Killing sbuild (pid=$main::sbuild_pid)\n");
+	kill( 15, $main::sbuild_pid );
+	$self->log("Waiting max. 2 minutes for sbuild to finish\n");
+	$SIG{'ALRM'} = sub ($) { die "timeout\n"; };
+	alarm( 120 );
+	eval "waitpid( $main::sbuild_pid, 0 )";
+	alarm( 0 );
+	if ($@) {
+	    $self->log("sbuild did not die!");
+	}
+	else {
+	    $self->log("sbuild died normally");
+	}
+	unlink( "SBUILD-REDO-DUMPED" );
+    }
+    unlink( $self->get('Config')->get('PIDFILE') );
+    $self->log("exiting now\n");
+    $self->close_log();
+    exit 1;
+}
+
 1;
