@@ -31,13 +31,16 @@ BEGIN {
 
     @ISA = qw(Exporter);
 
-    @EXPORT = qw(update upgrade distupgrade basesetup shell
-                 list_packages set_package_status);
+    @EXPORT = qw(update upgrade distupgrade clean autoclean autoremove basesetup
+                 shell list_packages set_package_status);
 }
 
 sub update ($$);
 sub upgrade ($$);
 sub distupgrade($$);
+sub clean ($$);
+sub autoclean ($$);
+sub autoremove ($$);
 sub basesetup ($$);
 sub shell ($$);
 sub list_packages ($$@);
@@ -79,6 +82,42 @@ sub distupgrade ($$) {
     return $?;
 }
 
+sub clean ($$) {
+    my $session = shift;
+    my $conf = shift;
+
+    $session->run_apt_command(
+	{ COMMAND => [$conf->get('APT_GET'), '-y', 'clean'],
+	  ENV => {'DEBIAN_FRONTEND' => 'noninteractive'},
+	  USER => 'root',
+	  DIR => '/' });
+    return $?;
+}
+
+sub autoclean ($$) {
+    my $session = shift;
+    my $conf = shift;
+
+    $session->run_apt_command(
+	{ COMMAND => [$conf->get('APT_GET'), '-y', 'autoclean'],
+	  ENV => {'DEBIAN_FRONTEND' => 'noninteractive'},
+	  USER => 'root',
+	  DIR => '/' });
+    return $?;
+}
+
+sub autoremove ($$) {
+    my $session = shift;
+    my $conf = shift;
+
+    $session->run_apt_command(
+	{ COMMAND => [$conf->get('APT_GET'), '-y', 'autoremove'],
+	  ENV => {'DEBIAN_FRONTEND' => 'noninteractive'},
+	  USER => 'root',
+	  DIR => '/' });
+    return $?;
+}
+
 sub basesetup ($$) {
     my $session = shift;
     my $conf = shift;
