@@ -81,7 +81,6 @@ sub init_allowed_keys {
 
     my %common_keys = (
 	'DISTRIBUTION'				=> {
-	    DEFAULT => 'unstable',
 	    SET => sub {
 		my $self = shift;
 		my $entry = shift;
@@ -92,6 +91,17 @@ sub init_allowed_keys {
 
 		my $override = ($self->get($key)) ? 1 : 0;
 		$self->set('OVERRIDE_DISTRIBUTION', $override);
+
+		#Now, we might need to adjust the MAILTO based on the
+		#config data. We shouldn't do this if it was already
+		#explicitly set by the command line option:
+		if (!$self->get('MAILTO_FORCED_BY_CLI') 
+		    && defined($self->get('DISTRIBUTION')) 
+		    && $self->get('DISTRIBUTION') 
+		    && $self->get('MAILTO_HASH')->{$self->get('DISTRIBUTION')}) {
+		    $self->set('MAILTO',
+		        $self->get('MAILTO_HASH')->{$self->get('DISTRIBUTION')});
+		}
 	    }
 	},
 	'OVERRIDE_DISTRIBUTION'			=> {
